/*=============================================================================
    Copyright (c) 2007-2008 Dan Marsden

    Distributed under the Boost Software License, Version 1.0. (See accompanying 
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/
#include <boost/detail/lightweight_test.hpp>
#include <boost/traversal/horizontal/all.hpp>

#include <vector>
#include <sstream>
#include <memory>
#include <boost/fusion/container/vector/vector10.hpp>
#include <boost/fusion/sequence/intrinsic/at.hpp>
#include <boost/optional.hpp>

using namespace boost;

namespace
{
    struct increment_always
    {
        template<typename T>
        bool operator()(T& t) const
        {
            ++t;
            return false;
        }
    };

    struct stream_always
    {
        explicit stream_always(std::ostream& stream)
            : stream(stream)
        {}

        template<typename T>
        bool operator()(T& t) const
        {
            stream << t;
            return false;
        }

        std::ostream& stream;
    };

    struct increment_once
    {
        template<typename T>
        bool operator()(T& t) const
        {
            ++t;
            return true;
        }
    };

    struct stream_once
    {
        explicit stream_once(std::ostream& stream)
            : stream(stream)
        {}

        template<typename T>
        bool operator()(T& t) const
        {
            stream << t;
            return true;
        }
        
        std::ostream& stream;
    };

    void all_tests()
    {
        int i = 0;
        traversal::all(i, increment_always());
        BOOST_TEST(i == 0);

        int* i_ptr = &i;
        traversal::all(i_ptr, increment_always());
        BOOST_TEST(i == 1);

        std::auto_ptr<int> a_ptr(new int(101));
        traversal::all(a_ptr, increment_always());
        BOOST_TEST(*a_ptr == 102);

        std::vector<int> std_v;
        std_v.push_back(101);
        std_v.push_back(202);
        std::ostringstream std_stream;
        traversal::all(std_v, stream_always(std_stream));
        BOOST_TEST(std_stream.str() == "101202");

        fusion::vector2<int,long> fusion_v(101, 202);
        std::ostringstream fusion_stream;
        traversal::all(fusion_v, stream_always(fusion_stream));
        BOOST_TEST(fusion_stream.str() == "101202");

        optional<int> o(101);
        traversal::all(o, increment_always());
        BOOST_TEST(o.get() == 102);
    }

    void once_tests()
    {
        int i = 0;
        traversal::all(i, increment_once());
        BOOST_TEST(i == 0);

        int* i_ptr = &i;
        traversal::all(i_ptr, increment_once());
        BOOST_TEST(i == 1);

        std::auto_ptr<int> a_ptr(new int(101));
        traversal::all(a_ptr, increment_once());
        BOOST_TEST(*a_ptr == 102);

        std::vector<int> std_v;
        std_v.push_back(101);
        std_v.push_back(202);
        std::ostringstream std_stream;
        traversal::all(std_v, stream_always(std_stream));
        BOOST_TEST(std_stream.str() == "101202");

        fusion::vector2<int,long> fusion_v(101, 202);
        std::ostringstream fusion_stream;
        traversal::all(fusion_v, stream_always(fusion_stream));
        BOOST_TEST(fusion_stream.str() == "101202");

        optional<int> o(101);
        traversal::all(o, increment_once());
        BOOST_TEST(o.get() == 102);
    }
}

int main()
{
    all_tests();
    once_tests();
    return boost::report_errors();
}
